/*-----------------------------------------------------------------------
			Copyright (c) Interactive Broadcasting 2005
	All Rights Reserved. No part of this software may be reproduced,
	transmitted, transcribed, stored in a retrieval system, or translated
	into any human or computer language, in any form or by any means,
	electronic, mechanical, magnetic, optical, manual or otherwise,
	without the express written permission of the copyright holder.
-----------------------------------------------------------------------*/

#include <vcl.h>
#pragma hdrstop

#include "xmlparser.h"

#include <sstream>

#include <cctype>
#include <cstdlib>

//---------------------------------------------------------------------------
#pragma package(smart_init)
//---------------------------------------------------------------------------

XMLParser::XMLParser(const std::string& f_name)
{
	file_name = f_name;
	if(file_name != "")
		file.open(file_name.c_str(),std::ios::in);
	if(!file)
	{
		std::ostringstream	buffer;
		buffer << "Unable to open file '" << f_name << "'";
		WriteError(buffer.str());
	}
}

XMLParser::~XMLParser()
{
	if(file)
		file.close();
}


void	XMLParser::BuildAttributes(const std::string& line)
{
	std::string::size_type	start = 0;
	std::string::size_type	end;
	attrib_list.clear();
	unsigned	len = line.length();
	if(len > 4)
		len -= 4;
	else
		return;

	AttribRec	rec;
	for(;;)
	{
		if(start >= len)
			return;
		end = line.find('=',start);
		if(end == std::string::npos)
			return;
		rec.name = line.substr(start,end - start);
		start = end + 2;
		end = line.find('\'',start);
		if(end == std::string::npos)
			return;
		rec.value = line.substr(start,end - start);
		start = end + 2;
		UnEscapeXML(rec.name);
		UnEscapeXML(rec.value);
		attrib_list.push_back(rec);
	}
}

void	XMLParser::BuildElement(const std::string& line)
{
	std::string::size_type	end = line.find(' ');
	if(end == std::string::npos) // no attributes
	{
		element = line;
		attrib_list.clear();
	}
	else
	{
		element = line.substr(0,end);
		BuildAttributes(line.substr(end + 1));
	}
}

std::string&	XMLParser::EscapeXML(std::string& text)
{
	std::ostringstream	buffer;
	int len = text.length();
	for(int count = 0;count < len;count++)
	{
		switch(text[count])
		{
			case '<'	:	buffer << "&lt;";			break;
			case '>'	:	buffer << "&gt;";			break;
			case '&'	:	buffer << "&amp;";		break;
			case '\'':	buffer << "&apos;";		break;
			case '\"':	buffer << "&quot;";		break;
			default	:	if(isascii(text[count]) == 0)
								buffer << '#';
							else
								buffer << text[count];
							break;
		}
	}
	text = buffer.str();
	return(text);
}

std::string	XMLParser::FindAttrib(const AttribList& attrib,
											const std::string& name,std::string& value)
{
	for(AttribList::const_iterator iter = attrib.begin();iter != attrib.end();iter++)
	{
		if((*iter).name == name)
		{
			value = (*iter).value;
			return(value);
		}
	}
	return(value = "");
}

bool	XMLParser::FindBoolAttrib(const AttribList& attrib,
											const std::string& name,bool default_val)
{
	std::string	temp;
	FindAttrib(attrib,name,temp);
	if(temp == "")
		return(default_val);
	else
		return(temp == "true");
}

int 	XMLParser::FindNumAttrib(const AttribList& attrib,
											const std::string& name,int default_val)
{
	std::string	temp;
	FindAttrib(attrib,name,temp);
	if(temp == "")
		return(default_val);
	else
		return(std::atoi(temp.c_str()));
}

bool	XMLParser::Parse()
{
	if(!file)
		return(false);

	std::string	line;
	while(std::getline(file,line))
	{
		if((line.substr(0,5).find("<?xml") == std::string::npos))	// skip XML header
			ProcessLine(line);
	}
	return(true);
}

void	XMLParser::Parse(const std::string& line)
{
	if((line.substr(0,5).find("<?xml") == std::string::npos))	// skip XML header
		ProcessLine(line);
}

void  XMLParser::ProcessLine(const std::string& line)
{
	std::string::size_type  blanks = line.find_first_not_of(" \t");
	if(blanks == std::string::npos)
		return;

	std::string::size_type	start;
	if(line[blanks] == '<')
		start = blanks;
	else
		start = std::string::npos;
	std::string::size_type	end = line.find('>');
	if((start != std::string::npos) &&(end != std::string::npos))
	{
		std::string	element_str(line.substr(start + 1, end - start - 1));
		if(element_str[0] == '/') 	//	it's an end element
		{
			EndElement(element_str.substr(1));
			return;
		}

		// element with just attributes
		if(element_str[element_str.length() - 1] == '/')
		{
			BuildElement(element_str.substr(0,element_str.length() - 1));
			StartElement(element,attrib_list);
			EndElement(element);
			return;
		}

		// element possibly also has text data
		BuildElement(element_str.substr(0,element_str.length()));
		StartElement(element,attrib_list);
		if((end + 1) >= line.length())
			return;
		std::string	line2(line.substr(end + 1));
		std::string::size_type	end2 = line2.find("</");
		if(end2 == std::string::npos)
		{
			UnEscapeXML(line2);
			TextData(line2);
			return;
		}
		else
		{
			std::string line4(line2.substr(0,end2));
			UnEscapeXML(line4);TextData(line4);
			std::string	line3(line2.substr(end2 + 2));
			std::string::size_type	end3 = line3.find('>');
			if(end2 != std::string::npos)
				EndElement(line3.substr(0,end3));
		}
	}
	else
	{
		if(end == std::string::npos)
		{
			std::string	the_line(line);
			UnEscapeXML(the_line);
			TextData(the_line);
		}
		else
		{
			std::string::size_type	start4 = line.find('<');
			std::string::size_type	end4 = line.find('>');
			if((start4 != std::string::npos) && (end4 != std::string::npos))
			{
				std::string	line5(line.substr(0,start4));
				UnEscapeXML(line5);
				TextData(line5);
				EndElement(line.substr(start4,end4-start4));
			}
			else
			{
				std::string	the_line(line);
				UnEscapeXML(the_line);
				TextData(the_line);
			}
		}
	}
}

std::string&	XMLParser::UnEscapeXML(std::string& text)
{
	std::ostringstream	buffer;
	int len = text.length();
	for(int count = 0;count < len;count++)
	{
		if(text[count] == '&')
		{
			switch(text[count + 1])
			{
				case 'a':
					if(text[count + 2] == 'm')	{ buffer << '&';	count += 4; }
					if(text[count + 2] == 'p')	{ buffer << '\'';	count += 5; }
					break;
				case 'g':	buffer << '>';		count += 3;	break;
				case 'l':	buffer << '<';		count += 3;	break;
				case 'q':	buffer << '\"';	count += 5;	break;
			}
		}
		else
				buffer << text[count];
	}
	text = buffer.str();
	return(text);
}







